from ..utils.session import Session
from ._base import Prompt


class LLaMA2Prompt(Prompt):
    """LLaMA Instruct prompt that follows this structure

    [INST] <<SYS>>
    {{ system_prompt }}
    <</SYS>>

    {{ user_msg_1 }} [/INST] {{ model_answer_1 }} [INST] {{ user_msg_2 }} [/INST]

    Args:
        system (str): system prompt

    """

    SYS_START = "<<SYS>>"
    SYS_END = "<</SYS>>"
    INST_START = "[INST]"
    TEXT_END = "[/INST]"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        """
        [INST] <<SYS>>
        {{ system_prompt }}
        <</SYS>>

        {{ user_msg_1 }} [/INST] {{ model_answer_1 }} [INST] {{ user_msg_2 }} [/INST]
        """

        prompt = f"{self.INST_START} {self.SYS_START}\n{self.system}\n{self.SYS_END}\n\n"
        for qa in session.history:
            # qa --> {"question": ..., "answer": ...}
            prompt += qa["question"] + " "  # type: ignore
            if qa["answer"] is None:
                break
            else:
                prompt += f"{self.TEXT_END} " + qa["answer"] + f" {self.INST_START} "

        if not prompt.endswith(self.TEXT_END):
            prompt += f"{self.TEXT_END}"

        return prompt


class TinyLLaMAPrompt(Prompt):
    """TinyLLaMa instruct prompt that follows this structure

    <|system|>
    You are a helpful AI assistant.</s>
    <|user|>
    How to explain Internet for a medieval knight?</s>
    <|assistant|>

    Args:
        system (str): system prompt

    """

    SYSTEM_START = "<|system|>"
    USER = "<|user|>"
    ASSISTANT = "<|assistant|>"
    TEXT_END = "</s>"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        """
        <|system|>
        You are a helpful AI assistant.</s>
        <|user|>
        How to explain Internet for a medieval knight?</s>
        <|assistant|>
        """

        prompt = f"{self.SYSTEM_START}\n{self.system}{self.TEXT_END}"
        for qa in session.history:
            # qa --> {"question": ..., "answer": ...}
            prompt += f"\n{self.USER}\n{qa['question']}{self.TEXT_END}"
            if qa["answer"] is None:
                break
            else:
                prompt += f"\n{self.ASSISTANT}\n{qa['answer']}{self.TEXT_END}"

        prompt += f"\n{self.ASSISTANT}\n"

        return prompt


class LLaMA3Prompt(Prompt):
    """LLaMA Instruct prompt that follows this structure

    <|begin_of_text|><|start_header_id|>system<|end_header_id|>

    {system_prompt}<|eot_id|><|start_header_id|>user<|end_header_id|>

    {prompt}<|eot_id|><|start_header_id|>assistant<|end_header_id|>

    Args:
        system (str): system prompt

    """

    PROMPT_START = "<|begin_of_text|>"
    START_HEADER = "<|start_header_id|>"
    END_HEADER = "<|end_header_id|>"
    TEXT_END = "<|eot_id|>"

    def __init__(self, system: str) -> None:
        self.system = system

    def prepare(self, session: Session) -> str:
        """Prepare prompt for model input

        Args:
            session (Session): dialog session

        Returns:
            str: model input prompt
        """

        """
        <|begin_of_text|><|start_header_id|>system<|end_header_id|>

        {system_prompt}<|eot_id|><|start_header_id|>user<|end_header_id|>

        {prompt}<|eot_id|><|start_header_id|>assistant<|end_header_id|>
        """

        prompt = f"{self.PROMPT_START} {self.START_HEADER}system{self.END_HEADER}\n\n{self.system}{self.TEXT_END}"
        for qa in session.history:
            # qa --> {"question": ..., "answer": ...}
            prompt += f"{self.START_HEADER}user{self.END_HEADER}\n\n{qa['question']}{self.TEXT_END}"
            prompt += f"{self.START_HEADER}assistant{self.END_HEADER}\n\n"
            if qa["answer"] is None:
                break
            else:
                prompt += qa["answer"] + self.TEXT_END

        return prompt



from typing import Dict, List, Optional, Tuple


class Session:
    """Session class.

    Args:
        questions (List[str], optional): questions. Defaults to [].
        answers (List[str], optional): answers. Defaults to [].
    """

    def __init__(
        self, questions: Optional[List[str]] = None, answers: Optional[List[str]] = None
    ) -> None:  # noqa: D107  # noqa: B006
        questions = questions or []
        answers = answers or []

        self.questions, self.answers = self._load(questions, answers)
        self.q_reset_len = len(self.questions)
        self.a_reset_len = len(self.answers)

    def _load(self, questions: List[str], answers: List[str]) -> Tuple[List[str], List[str]]:
        """Load questions and answers.

        Args:
            questions (List[str]): questions
            answers (List[str]): answers

        Returns:
            Tuple[List[str], List[str]]: valid questions and answers
        """
        assert isinstance(questions, list), f"questions must be a list. Got {type(questions)} instead."

        assert isinstance(answers, list), f"answers must be a list. Got {type(answers)} instead."

        if len(questions) < len(answers):
            print(
                f"[WARNING] Found {len(answers) - len(questions)} more answers than questions. Extra {len(answers) - len(questions)} answers will be discarded."
            )
            up_to = len(questions)
        elif len(answers) < len(questions) - 1:
            print(
                f"[WARNING] Found {len(questions) - len(answers)} more questions than answers. Extra {len(questions) - len(answers) - 1} questions will be discarded."
            )
            up_to = len(answers) + 1
        else:
            up_to = len(questions)

        questions = questions[:up_to]
        answers = answers[: up_to - 1] if len(answers) == up_to - 1 else answers[:up_to]

        return questions, answers

    def reset(self) -> None:
        """Reset conversation."""
        self.questions = self.questions[: self.q_reset_len]
        self.answers = self.answers[: self.a_reset_len]

    def add_question(self, question: str) -> None:
        """Add question to conversation.

        Args:
            question (str): question

        Raises:
            ValueError: if there is nos answer for the previous question
        """
        if len(self.questions) != len(self.answers):
            raise ValueError("Cannot add question. There is no answer for the previous question.")
        self.questions.append(question)

    def add_answer(self, answer: str) -> None:
        """Add answer to conversation.

        Args:
            answer (str): answer

        Raises:
            ValueError: if there is no question for the answer
        """
        if len(self.questions) == len(self.answers):
            raise ValueError("Cannot add answer. There is no question for the answer.")
        self.answers.append(answer)

    @property
    def history(self) -> List[Dict[str, Optional[str]]]:
        """Conversation history.

        Returns:
            List[Dict[str, Optional[str]]]: session history in the form of a list of dictionaries with keys "question" and "answer"
        """
        history = []
        for i in range(len(self.questions)):
            q = self.questions[i]
            a = self.answers[i] if i < len(self.answers) else None
            history.append({"question": q, "answer": a})
        return history

    def to_string(self, last_k: int = 0) -> str:
        """String representation of conversation.

        Args:
            last_k (int): number of questions and answers to consider. If last_k==0, all questions and answers are considered. Defaults to 0.

        Returns:
            str: conversation
        """

        assert isinstance(last_k, int), f"last_k must be an int. Got {type(last_k)} instead."

        if last_k > len(self.history):
            print(
                f"[WARNING] last_k={last_k} is greater than the number of answers ({len(self.history)}). last_k will be set to {len(self.history)}."
            )
            last_k = 0  # all history

        _history = self.history[-last_k:]
        out = ""
        for qa in _history:
            out += f"Question: {qa['question']}\n"
            if qa["answer"] is not None:
                out += f"Answer: {qa['answer']}\n"
        return out

    def __len__(self) -> int:
        """Length of conversation.

        Returns:
            int: length of conversation
        """
        return len(self.questions)

    def __getitem__(self, idx: int) -> Dict[str, Optional[str]]:
        """Get item from conversation.

        Args:
            idx (int): index

        Returns:
            Dict[str, Optional[str]]: item
        """
        return self.history[idx]



import os
from pathlib import Path
from typing import Dict, Generator, List, Union

import mlx.core as mx
import numpy as np
import torch
from safetensors import safe_open
from tqdm import tqdm

from .time import Timing


def save_weights(weights: Dict[str, mx.array], output_path: str) -> None:
    """Save MLX weights to a given path.

    Args:
        weights (Dict[str, mx.array]): MLX weights dict (key, mx.array)
        output_path (str): path to save weights
    """
    output_dir = os.path.dirname(output_path)
    if len(output_dir) > 0:
        os.makedirs(output_dir, exist_ok=True)
    if output_path.endswith(".safetensors"):
        mx.save_safetensors(output_path, weights, metadata={"format": "mlx"})
    else:
        mx.savez(output_path, **weights)


def smart_load(ckpt_paths: Union[List[str], str]) -> Generator:
    """Load a checkpoint from a list of paths.

    Args:
        ckpt_paths (Union[List[str], str]): path to checkpoint or multiple paths

    Yields:
        Generator: a generator of state_dict
    """

    if isinstance(ckpt_paths, str):
        ckpt_paths = [ckpt_paths]

    for ckpt_path in ckpt_paths:
        if ckpt_path.endswith(".safetensors"):
            state_dict = mx.load(ckpt_path)
        else:
            state_dict = torch.load(ckpt_path, map_location="cpu")
        yield state_dict


def weights_to_mlx(
    ckpt_paths: Union[List[str], str],
    show_kv: bool = False,
    tensor_type: torch.dtype = torch.float16,
) -> Dict[str, mx.array]:
    """Convert a checkpoint from PyTorch or safetensors to a MLX weights (Dict[str, mx.array]).

    Args:
        ckpt_paths (Union[List[str], str]): path to checkpoint
        show_kv (bool, optional): whether to show key-value pairs. Defaults to False.
        tensor_type (torch.dtype, optional): tensor type. Defaults to torch.float16.

    Returns:
        Dict[str, mx.array]: MLX weights dict (key, mx.array)
    """
    weights = {}
    for state_dict in smart_load(ckpt_paths):
        for k, w in state_dict.items():
            if show_kv:
                print(k, w.shape)
            if isinstance(w, torch.Tensor):
                w = w.to(tensor_type).numpy()
                w = mx.array(w)
            weights[k] = w
    return weights
